/*
 *   Copyright (C) 2008-2009 Anders Aagaard <aagaande@gmail.com
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License version 2 as
 *   published by the Free Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <KStandardDirs>

#include "database.h"

Database::Database(FSSettings *settings)
{
    qDebug() << "Creating database object";
    transaction_ = false;
    treatSpaceAsStar = settings->treatSpaceAsStar;
    searchTermStrict = settings->searchTermStrict;
    searchFieldName = settings->searchFieldName;

    if (!database_initialized) {
        qDebug() << "Opening database - new database connection";
        database_initialized = true;

        KStandardDirs *dirs = KGlobal::dirs();
        QDir saveLocation = QDir(dirs->saveLocation("data", "fsrunner", true));
        assert(saveLocation.exists());

        connection_ = QSqlDatabase::addDatabase("QSQLITE", CONNECTIONNAME);

        QFile file(saveLocation.absoluteFilePath("database.sql"));
#ifdef CONVERTSETTINGS

        if (settings->dbversion < 2) {
            if (file.exists()) {
                qWarning("Database outdated, rescanning.");
                QMessageBox::information(0, "FSRunner - Database", "Database outdated, rescanning.");
                file.remove();
            }
            saveVersion("dbversion",2);
            settings->dbversion = 2;
        }
#endif

        connection_.setDatabaseName(saveLocation.absoluteFilePath("database.sql"));
        if (!connection_.open()) {
            QMessageBox::critical(0, "FSRunner - Database", QString("Failed opening database, error : %1\nDid you install the qt sqlite3 driver?").arg(connection_.lastError().text()));
            qCritical("Error opening database %s", qPrintable(connection_.lastError().text()));
            return;
        }
        assert(connection_.driver()->hasFeature(QSqlDriver::LastInsertId));
        createTables();
        cleanDatabase(settings);
    } else {
        qDebug() << "Opening database - reusing database connection";
    }

    assert(isValid());
}

bool Database::isValid()
{
    bool isValid = connection_.isValid();
    assert(!isValid || connection_.tables().contains("Entries"));
    return isValid;
}

Database::~Database()
{
    qDebug() << "Closing database";
    forceCommit();
    assert(!transaction_);
    QMutexLocker locker(&mutex_);
}

void Database::createTables()
{
    QString id = "id INTEGER NOT NULL PRIMARY KEY";
    if (!connection_.tables().contains("Entries")) {
        /*
        Entries Table layout:
        id - primary key
        name TEXT - path.dirName() || path.fileName()
        path TEXT - abolute path
        priority INTEGER - deals with sorting, every time an entry is selected priority is increased by one, so popular entries are sorted first
        parent INTEGER - refers Entries.Id where path.cdup == parent.path, could do (SELECT id FROM Entries WHERE path=path.cdup()) every time, but cached for performance reasons.
          */
        QSqlQuery query(connection_);
        QUERY_EXEC(query.prepare(QString("CREATE TABLE Entries (%1, name TEXT NOT NULL, absolutePath TEXT NOT NULL, priority INTEGER NOT NULL, parent INTEGER NOT NULL, isdirectory BOOL NOT NULL)").arg(id)));
        QUERY_EXEC(query.exec());
    }
}

// Remove every path tree from the database that isn't in the configuration file.
void Database::cleanDatabase(FSSettings *settings)
{
    QMutexLocker locker(&mutex_);
    endTransaction(); // Force dumping transaction to disk before cleanup.
    beginTransaction();
    // Get paths where parent=0 in database

    // Get paths we are supposed to scan.
    QStringList paths;
    for (int i = 0; i < settings->folders.size(); i++)
        paths << settings->folders.at(i).path;

    QList<int> removableIds;
    QSqlQuery query(connection_);
    QUERY_EXEC(query.prepare("SELECT id,absolutePath FROM Entries WHERE parent=0"));
    QUERY_EXEC(query.exec());
    while (query.next()) {
        int id = query.value(0).toInt();
        QString absolutePath = query.value(1).toString();
        if (!paths.contains(absolutePath)) {
            qDebug() << "Cleaning" << absolutePath << "from database";
            removableIds << id;
        }
    }
    locker.unlock();
    remove(removableIds);
    locker.relock();
    endTransaction();
}

// Remove a list of id's (primary key's)
void Database::remove(QList<int> ids)
{
    if (ids.size() == 0)
        return;
    qDebug() << "Removing id's :" << ids << ids.size();
#ifdef DEBUG
    for (int i = 0; i < ids.size(); i++)
        assert(ids.at(i) > 0);
#endif
#ifdef DEBUG_NO_CHANGE
    assert(!DEBUG_NO_CHANGE);
#endif
    QMutexLocker locker(&mutex_);
    beginTransaction();

    // First select all id's that has us as parents (aka our children)
#ifdef DEBUG
    QString queryString = "SELECT id,name FROM Entries WHERE parent IN (";
#else
    QString queryString = "SELECT id FROM Entries WHERE parent IN (";
#endif
    for (int i = 0; i < ids.size(); i++) {
        if (ids.at(i) > 0) {
            queryString.append(QString::number(ids.at(i)));
            if (ids.size() - 1 != i)
                queryString.append(",");
        }
    }
    queryString.append(")");
    qDebug() << queryString;
    QSqlQuery query(connection_);
    QUERY_EXEC(query.prepare(queryString));
    QUERY_EXEC(query.exec());

    // Convert this to a list of removeable id's (all our children), and remove them (recursive function)
    QList<int> removableIds;
    while (query.next()) {
        removableIds << query.value(0).toInt();
#ifdef DEBUG
        qDebug() << "   removing sub entry" << query.value(1).toString() << query.value(0).toInt();
#endif
    }
    locker.unlock();
    remove(removableIds);
    locker.relock();

    // And now remove "ourselfes":
    queryString = "DELETE FROM Entries WHERE id IN (";
    for (int i = 0; i < ids.size(); i++) {
        if (ids.at(i) > 0) {
            queryString.append(QString::number(ids.at(i)));
            if (ids.size() - 1 != i)
                queryString.append(",");
        }
    }
    queryString.append(")");
    qDebug() << queryString;
    QUERY_EXEC(query.prepare(queryString));
    QUERY_EXEC(query.exec());

#ifdef DEBUG
    QUERY_EXEC(query.exec("SELECT * FROM Entries WHERE parent NOT IN (select id from entries) AND parent > 0"));
    assert(!query.next());
#endif
}

void Database::remove(QString absolutePath)
{
#ifdef DEBUG_NO_CHANGE
    assert(!DEBUG_NO_CHANGE);
#endif
    qDebug() << "remove" << absolutePath;
    QMutexLocker locker(&mutex_);

    QSqlQuery query(connection_);
    QUERY_EXEC(query.prepare("SELECT id FROM Entries WHERE absolutePath=?"));
    query.addBindValue(absolutePath);
    QUERY_EXEC(query.exec());

    if (!query.next())
        return;

    QList<int> removableIds;
    removableIds << query.value(0).toInt();
    assert(!query.next());
    locker.unlock();
    remove(removableIds);

}

void Database::beginTransaction()
{
    assert(mutex_.tryLock() == false);
    if (transaction_)
        return;
    assert(isValid());
    qDebug() << "starting database transaction";
    if (!connection_.transaction()) {
        QMessageBox::critical(0, "FSRunner - Database", QString("Error starting transaction: %1").arg(connection_.lastError().text()));
        qFatal("Error starting transaction: %s", qPrintable(connection_.lastError().text()));
    }
    transaction_ = true;
}

void Database::endTransaction()
{
    assert(mutex_.tryLock() == false);
    if (!transaction_)
        return;
    assert(isValid());
    qDebug() << "committing transaction to database";
    if (!connection_.commit()) {
        QMessageBox::critical(0, "FSRunner - Database", QString("Error ending transaction: %1").arg(connection_.lastError().text()));
        qFatal("Error comitting transaction: %s", qPrintable(connection_.lastError().text()));
    }
    transaction_ = false;
}

void Database::forceCommit()
{
    qDebug() << "forceCommit";
    QMutexLocker locker(&mutex_);
    endTransaction();
}

const QHash<QString, int> Database::getEntries(const int parent)
{
    assert(parent > 0);
    qDebug() << "getEntries" << parent;
    QMutexLocker locker(&mutex_);
    QSqlQuery query(connection_);
    QUERY_EXEC(query.prepare("SELECT id,name FROM Entries WHERE parent=?"));
    query.addBindValue(parent);
    QUERY_EXEC(query.exec());

    QHash<QString, int> entries;
    while (query.next()) {
        int id = query.value(0).toInt();
        QString name = query.value(1).toString();
        qDebug() << "   " << id << name;
        assert(!entries.contains(name));
        entries[name] = id;
    }
    qDebug() << "getEntries found " << entries.size() << "entries";
    return entries;
}
#ifndef LOWMEM_SCAN
const QVarLengthArray<QHash<QString, int> > Database::getEntries()
{
    qDebug() << "getEntries";
    QMutexLocker locker(&mutex_);
    int maxSize;
    {
        QSqlQuery query(connection_);
        QUERY_EXEC(query.prepare("SELECT max(parent) FROM Entries"));
        QUERY_EXEC(query.exec());
        QUERY_EXEC(query.next());
        maxSize = query.value(0).toInt();
    }
    QVarLengthArray<QHash<QString, int> > entries(maxSize + 1);
    QSqlQuery query(connection_);
    QUERY_EXEC(query.prepare("SELECT id,name,parent FROM Entries WHERE parent > 0"));
    QUERY_EXEC(query.exec());
    while (query.next()) {
        int id = query.value(0).toInt();
        QString name = query.value(1).toString();
        int parent = query.value(2).toInt();
        assert(entries.size() > parent);
        assert(!entries[parent].contains(name));
        entries[parent][name] = id;
    }
    return entries;
}
#endif

int Database::addEntry(QString name, QString absolutePath, bool directory, int priority, int parent)
{
    qDebug() << "AddEntry" << absolutePath << name << parent;
#ifdef DEBUG_NO_CHANGE
    assert(!DEBUG_NO_CHANGE);
#endif
    QMutexLocker locker(&mutex_);
    beginTransaction();

    QSqlQuery query(connection_);
    QUERY_EXEC(query.prepare("INSERT INTO Entries VALUES(NULL, ?, ?, ?, ?, ?)"));
    query.addBindValue(name);
    query.addBindValue(absolutePath);
    // Directories have higher prio than files:
    if (directory)
        query.addBindValue(priority + 5);
    else
        query.addBindValue(priority);
    query.addBindValue(parent);
    query.addBindValue(directory);
    QUERY_EXEC(query.exec());
    return query.lastInsertId().toInt();
}

int Database::getEntry(QString absolutePath)
{
    qDebug() << "getEntry" << absolutePath;
    QMutexLocker locker(&mutex_);

    // First try to get id
    QSqlQuery query(connection_);
    QUERY_EXEC(query.prepare("SELECT id FROM Entries WHERE absolutePath=?"));
    query.addBindValue(absolutePath);
    QUERY_EXEC(query.exec());
    if (query.next()) {
        int id = query.value(0).toInt();
        assert(!query.next());
        return id;
    }
    return -1;
}

const QList<Plasma::QueryMatch> Database::match(Plasma::AbstractRunner *runner, QString matchString)
{
    static QReadWriteLock countLocker;
    static int counter = 0;
    int count;
    {
        // For every loop, set count = current, current +1, and check count == current in the loop to make sure we are the latest scanner.
        QWriteLocker locker(&countLocker);
        count = ++counter;
    }

    assert(runner);
    QMutexLocker locker(&mutex_);
    endTransaction();

    QString m = matchString;
    m.replace("*", "%");
    if (treatSpaceAsStar)
        m.replace(" ", "%");
    if (searchTermStrict)
        m = QString("%1%").arg(m);
    else
        m = QString("%%1%").arg(m);
    qDebug() << "SearchTerm" << searchTermStrict << m;

    QSqlQuery query(connection_);
    if (searchFieldName) {
        QUERY_MATCH(query.prepare("SELECT name,absolutePath,priority FROM entries WHERE name LIKE ? ORDER BY priority"));
    } else {
        QUERY_MATCH(query.prepare("SELECT name,absolutePath,priority FROM entries WHERE absolutePath LIKE ? ORDER BY priority"));
    }
    query.addBindValue(m);
    QUERY_MATCH(query.exec());

    QList<Plasma::QueryMatch> matches;
    while (query.next()) {
        Plasma::QueryMatch match(runner);
        QString name = query.value(0).toString();
        QString absolutePath = query.value(1).toString();
        double priority = query.value(2).toDouble();
        qDebug() << "   Adding" << name << absolutePath << priority;
        if (name == matchString)
            match.setType(Plasma::QueryMatch::ExactMatch);
        else
            match.setType(Plasma::QueryMatch::PossibleMatch);

        // Deal with things that arnt threadsafe:
        Plasma::AbstractRunner::bigLock()->lock();
        KUrl url;
        url.setPath(absolutePath);
        match.setIcon(KIcon(KMimeType::iconNameForUrl(url)));
        Plasma::AbstractRunner::bigLock()->unlock();

        match.setData(absolutePath);
        match.setText(name);
        match.setRelevance(1 / (priority + 1));
        match.setSubtext(QFileInfo(absolutePath).absolutePath());
        matches << match;
        {
            QReadLocker locker(&countLocker);
            if (count != counter) {
                // We are no longer the latest match scanner.
                qDebug() << "Stopping scan in progress and returning";
                return matches;
            }
        }
    }
    qDebug() << "Found" << matches.size() << "for query" << m << "converted from" << matchString;
    return matches;
}
