/*
 *   Copyright (C) 2008-2009 Anders Aagaard <aagaande@gmail.com
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License version 2 as
 *   published by the Free Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "dirwatcher.h"

DirectoryLookup::DirectoryLookup()
{}

void DirectoryLookup::addFolder(Folder f)
{
    QWriteLocker locker(&lock_);
    assert(!folders_.contains(f.path));
    folders_[f.path] = f;
}

void DirectoryLookup::getSettings(const QString &path,  int &subdirs, int &priority, bool &files, bool &slow)
{
    QReadLocker locker(&lock_);
    if (folders_.contains(path)) {
        subdirs = folders_[path].subdirs;
        priority = folders_[path].priority;
        files = folders_[path].files;
        slow = folders_[path].slow;
    }
}

Folder DirectoryLookup::lookupSettings(const QString &pathString)
{
    QDir path(pathString);
    path.makeAbsolute();
    assert(path.exists());

    QReadLocker locker(&lock_);
    // Scan up in the path, to find out what settings we should have for this directory.
    int i = 0; // Dirs to substract from folder.subdirs, aka directories we have walked to find it.

    bool ok = true;
    while (ok) {
        if (folders_.contains(path.absolutePath())) {
            // Found it!
            Folder f = folders_.value(path.absolutePath());
            f.path = pathString;
            f.subdirs = f.subdirs - i;

            return f;
        }
        i++;
        ok = path.cdUp();
    }

    QMessageBox::critical(0, "FSRunner - DirWatcher", QString("Failed to walk up to find scan folder %1 %2").arg(pathString).arg(path.absolutePath()));
    qFatal("Failed to walk up to find scan folder %s %s", qPrintable(pathString), qPrintable(path.absolutePath()));
    // This exists to silence compile errors.
    return Folder();
}

DirWatcher::DirWatcher(FSSettings *settings, Database *database, QObject *parent) : QThread(parent)
{
    database_ = database;
    stopScan = false;

    bool ok = connect(&dirwatch_, SIGNAL(deleted(const QString&)), SLOT(deleted(const QString&)));
    assert(ok);
    ok = connect(&dirwatch_, SIGNAL(dirty(const QString&)), SLOT(dirty(const QString&)));
    assert(ok);

    loadFolders(settings);
}

void DirWatcher::stop()
{
    qDebug() << "DirWatcher - Stop called, triggering abort";
    {
        QWriteLocker locker(&scanningLock);
        scanList.clear();
        stopScan = true;
    }
    wait();
    qDebug() << "DirWatcher -- All scans aborted";
}

void DirWatcher::loadFolders(FSSettings *settings)
{
#ifndef LOWMEM_SCAN
    useEntriesCache_ = true;
#endif
    folderLookup_.clear();
    for (int i = 0; i < settings->folders.size(); i++) {
        //assert(!folders.contains(loadedFolders.at(i).path));
        directoryLookup_.addFolder(settings->folders.at(i));
        int id = database_->getEntry(settings->folders.at(i).path);
        if (id == -1)
            id = database_->addEntry(QDir(settings->folders.at(i).path).dirName(), settings->folders.at(i).path, true, settings->folders.at(i).priority, 0);
        folderLookup_[settings->folders.at(i).path] = id;
        scan(settings->folders.at(i));
    }
}

void DirWatcher::scanPath(const QString &pathString, int subdirs, int priority, bool files, bool slow, int parent)
{
    QDir path(pathString);
    path.makeAbsolute();
    assert(path.exists());
    QString absolutePath = path.absolutePath();

    // Check if anything should override our subdirs/priority etc:
    directoryLookup_.getSettings(absolutePath, subdirs, priority, files, slow);
    if (subdirs < 0)
        return;

    // Begin actual scan of data:
    QMutexLocker locker(&mutex_);
    qDebug() << "Scanning" << path << subdirs << parent;

    // If we arnt already monitoring this directory we should be.
#ifdef DEBUG
    QFileInfo tmpDebug(absolutePath);
    if (!tmpDebug.isDir())
        qFatal("Directory not directory : %s", qPrintable(absolutePath));
#endif
    if (!dirwatch_.contains(absolutePath))
        dirwatch_.addDir(absolutePath, KDirWatch::WatchDirOnly);

    // Setup directory filter:
    int dirFilter = QDir::NoDotAndDotDot;
    if (files)
        dirFilter = dirFilter | QDir::AllEntries;
    else
        dirFilter = dirFilter | QDir::Drives | QDir::Dirs;

    // Note, before changing this to follow symlinks, directory lookup system has to be changed.
    // Folders have to be added every time we find a new link, so that we can do reverse lookup.
    dirFilter = dirFilter | QDir::NoSymLinks;
    path.setFilter((QDir::Filter)dirFilter);

    if (slow)
#ifdef LOWMEM_SCAN
        msleep(200);
#else
        msleep(30);
#endif

#ifdef LOWMEM_SCAN
    // Get list of entries in the database:
    QHash<QString, int> databaseEntries = database_->getEntries(parent); // Scan removes all data we should keep, and only invalid entries are left.
#else
    QHash<QString, int> databaseEntries;
    if (useEntriesCache_) {
        if (entriesCache_.size() > parent)
            databaseEntries = entriesCache_[parent];
    } else {
        databaseEntries = database_->getEntries(parent);
    }

#endif

    QFileInfoList list = path.entryInfoList();
    for (int i = 0; i < list.size(); ++i) {
        CHECK_STOPSCAN;
        QFileInfo fileInfo = list.at(i);
        int dbId = databaseEntries.take(fileInfo.fileName());
        assert(!fileInfo.isSymLink());
        if (fileInfo.isDir()) {
            if (!dirwatch_.contains(fileInfo.absoluteFilePath())) {
                // This is a directory we haven't scanned before.
                if (dbId == 0) {
                    // dbId = 0, we're not in the database with the same parent, try getting it from the folder Lookup (this'll work when for example having seperate folder entries for $HOME and $HOME/Desktop
                    dbId = folderLookup_.value(fileInfo.absoluteFilePath());
                    if (dbId == 0) {
                        // Ok, it's not in folderLookup, add the entry.
                        assert(database_->getEntry(fileInfo.absoluteFilePath()) == -1);
                        dbId = database_->addEntry(fileInfo.fileName(), fileInfo.absoluteFilePath(), true, priority, parent);
                    }
                }
                locker.unlock();
                scanPath(fileInfo.absoluteFilePath(), subdirs - 1, priority + 10, files, slow, dbId);
                locker.relock();
            }
        } else {
            // Check if we should ignore this file
	    // TODO : Does this parse .suffix every time?
            if ((fileInfo.suffix() == "bak") || (fileInfo.fileName().rightRef(1) == "~") || (fileInfo.suffix() == "old") || fileInfo.fileName() == "lost+found")
                continue;

            // Add file to database if it doesn't exist
            assert(files);
            if (dbId == 0)
                database_->addEntry(fileInfo.fileName(), fileInfo.absoluteFilePath(), false, priority, parent);
        }
    }
    qDebug() << "Directory scan completed";

    // Deal with invalid data
    QList<int> removableIds;
    QList<int> folderLookupIds = folderLookup_.values();
    QHash<QString, int>::const_iterator i = databaseEntries.constBegin();
    while (i != databaseEntries.constEnd()) {
        if (!folderLookupIds.contains(i.value())) {
            qDebug() << "   Removing:" << i.key() << i.value();
            removableIds << i.value();
        }
        ++i;
    }
    database_->remove(removableIds);
}

void DirWatcher::run()
{
    ioprio();
    forever {
        Folder f;
        {
            QWriteLocker locker(&scanningLock);
            if (scanList.size() == 0)
                break;
            f = scanList.dequeue();
        }
        int parent = database_->getEntry(f.path);
        if (parent < 1) {
            qDebug() << "WARNING, couldn't get parent for" << f.path;
            assert(0);
        }

#ifndef LOWMEM_SCAN
        if (useEntriesCache_)
            entriesCache_ = database_->getEntries();
#endif

        QDir path(f.path);
        if (path.exists())
            scanPath(f.path, f.subdirs, f.priority, f.files, f.slow, parent);
    }

    database_->forceCommit();
#ifndef LOWMEM_SCAN
    entriesCache_.clear();
    useEntriesCache_ = false;
#endif
}

void DirWatcher::scan(Folder folder)
{
    qDebug() << "Queueing scan of folder" << folder.path;
    assert(!folder.path.isEmpty());
    {
        // Check if path already exists, aka haven't been scanned yet.
        QReadLocker locker(&scanningLock);
        for (int i = 0; i < scanList.size(); i++)
            if (folder.path == scanList.at(i).path)
                return;
    }
    {
        // No, add to path
        QWriteLocker locker(&scanningLock);
        scanList.append(folder);
    }
    // Start scan.

    if (!isRunning())
        start(IdlePriority);
}

void DirWatcher::deleted(const QString &path)
{
    qDebug() << "Path deleted" << path;
    QFileInfo fileInfo(path);

    QMutexLocker locker(&mutex_);
    assert(dirwatch_.contains(path));
    dirwatch_.removeDir(path);

    database_->remove(fileInfo.absoluteFilePath());
    database_->forceCommit();
}

void DirWatcher::dirty(const QString &path)
{
    // Workaround : Kde 4.2.67 seems to link to the file that's dirty even when dirmode is set., and it returns the path twice, once with trailing slash and once without.
    QDir x(path);
    {

        if (!x.exists())
            return;
    }
    scan(directoryLookup_.lookupSettings(x.absolutePath()));
}

