/*
 *   Copyright (C) 2008-2009 Anders Aagaard <aagaande@gmail.com
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License version 2 as
 *   published by the Free Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifndef DirWatcher_H
#define DirWatcher_H

#include <QObject>
#include <QThread>
#include <QDir>
#include <QHash>
#include <QReadWriteLock>
#include <kdirwatch.h>

#include "database.h"
#include "config.h"
#include "ioprio.h"


#define CHECK_STOPSCAN { \
        QReadLocker locker(&scanningLock); \
        if (stopScan) \
            return; \
        }

/*!
  Class for doing reverse directory lookup.
  This is needed when a directory is marked dirty, to check what settings we should use to scan it.
  */
class DirectoryLookup
{
private:
    // This is the folders we are currently monitoring:
    QReadWriteLock lock_;
    QHash<QString, Folder> folders_;
public:
    DirectoryLookup();
    void addFolder(Folder f);

    /*!
      This function allows a path to override settings the lower it is in the tree.
      For example settings for $HOME/ will override /
      */
    void getSettings(const QString &path,  int &subdirs, int &priority, bool &files, bool &slow);

    /*!
      This allows setting lookup for a subpath, for example lookup $HOME/tmp/directory to $HOME settings.
      */
    Folder lookupSettings(const QString &path);
};

/*!
  folders - Configuration file folders, what we are supposed to scan.
  scanList - List of folders we haven't scanned yet, rescans due to dirty directories are also queue'd here.
  scanPath - Recursive function to scan a tree
  scan - Function used to queue a new scan folder
  */
class DirWatcher : public QThread
{
    Q_OBJECT
private:
    QMutex mutex_; // Lock for the scanning process itself.
    Database *database_;
    KDirWatch dirwatch_;
    DirectoryLookup directoryLookup_;
    QHash<QString, int> folderLookup_; // This holds the main folders in the database.
#ifndef LOWMEM_SCAN
    QVarLengthArray<QHash<QString, int> > entriesCache_;
    bool useEntriesCache_; // Only use this on the "main" scans, not on updates.
#endif

    // This is the folders we are currently scanning (can be more than above, due to rescanning dirty paths)
    QReadWriteLock scanningLock;
    bool stopScan;
    QQueue<Folder> scanList;

    void scan(Folder folder); // Queue a scan
    void scanPath(const QString &path, int subdirs, int priority, bool files, bool slow, int parent); // Recursive function to scan a tree
    void loadFolders(FSSettings *settings); // Load folders from configuration file, populate folders, and queue scan's.
public:
    DirWatcher(FSSettings *settings, Database *database, QObject *parent);
    virtual void run();
    void stop();
private slots:
    void deleted (const QString &path);
    void dirty (const QString &path);
};
#endif
