/*
 *   Copyright (C) 2008-2009 Anders Aagaard <aagaande@gmail.com
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License version 2 as
 *   published by the Free Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "fs.h"

#include <KApplication>
#include <KIcon>

static const int MAX_ADDED_ACTIONS = 3;
static const int KRUNNER_HEIGHT = 600;
static const int KRUNNER_WEIGHT = 800;
static const QString KRUNNER_CONFIG_FILE = "/home/thibault/.kde/share/config/krunnerrc";

FSRunner::FSRunner(QObject* parent, const QVariantList &args)
    : Plasma::AbstractRunner(parent, args)
{
    qDebug() << "FSRunner initializing";
    Q_UNUSED(args)
    KGlobal::locale()->insertCatalog("krunner_fsrunner");
    setObjectName("FS");
    //setSpeed(AbstractRunner::SlowSpeed); // We're not all that slow.
    setPriority(LowPriority);

    KGlobal::locale()->insertCatalog("libkonq");
    KGlobal::locale()->insertCatalog("krunner_fsrunner");
    actions_ = new KonqMenuActions;

    Plasma::AbstractRunner::bigLock()->lock();
    FSSettings settings = loadConfig();
    enableActions_=settings.enableActions;
    database_ = new Database(&settings);
    if (!database_->isValid()) {
        qCritical("Database invalid!, dying gracefully");
        delete database_;
        database_ = 0;
        dirWatcher_ = 0;
    } else {
        dirWatcher_ = new DirWatcher(&settings, database_, this);
    }
    Plasma::AbstractRunner::bigLock()->unlock();
}

void FSRunner::reloadConfiguration()
{
    Plasma::AbstractRunner::bigLock()->lock();
    qDebug() << "Reloading configuration - Wait for scan halt";
    dirWatcher_->stop();
    qDebug() << "Reloading configuration - Freeing dirwatcher";
    delete dirWatcher_;

    KGlobal::locale()->insertCatalog("libkonq");
    KGlobal::locale()->insertCatalog("krunner_fsrunner");
    
    FSSettings settings = loadConfig();
    enableActions_=settings.enableActions;

    if (database_ != 0) {
        qDebug() << "Reloading configuration - Freeing database";
        delete database_;
    } else {
        qDebug() << "Reloading configuration - Freeing database - not needed, db already NULL";
    }
    qDebug() << "Reloading configuration - Reloading database";
    database_ = new Database(&settings);
    if (!database_->isValid()) {
        qCritical("Database invalid!, dying gracefully");
        delete database_;
        database_ = 0;
        dirWatcher_ = 0;
        Plasma::AbstractRunner::bigLock()->unlock();
        return;
    }

    qDebug() << "Reloading configuration - Reloading dirwatcher";
    dirWatcher_ = new DirWatcher(&settings, database_, this);
    Plasma::AbstractRunner::bigLock()->unlock();
}

FSRunner::~FSRunner()
{
    qDebug() << "fsrunner shutting down";
    dirWatcher_->stop();
    delete database_;
    delete actions_;
}

void FSRunner::match(Plasma::RunnerContext &context)
{
    if (context.query().size() < 2)
        return;
    qDebug() << "Looking for query : " << context.query();

    // Seperate database object per thread
    FSSettings settings = loadConfig();
    Database *threadDb = new Database(&settings);
    assert(threadDb->isValid());
    context.addMatches(context.query(), threadDb->match(this, context.query()));
    delete threadDb;
}

void FSRunner::run(const Plasma::RunnerContext &context, const Plasma::QueryMatch &match)
{
if(enableActions_) {
    // If we can use the action system we will:
    if (QAction *a = match.selectedAction()) {
        if (a != action("open")) {
            qDebug() << "Triggering action" << context.query() << match.data().toString();
            match.selectedAction()->trigger();
            return;
        }
    }
}

    QString absolutePath = match.data().toString();
    qDebug() << "Running" << context.query() << absolutePath;
    (void)new KRun(absolutePath, 0);
}


// The following code is shared with the nepomuksearch runner and written by Ryan P. Bitanga <ryan.bitanga@gmail.com>.
QList<QAction*> FSRunner::parseMenu(QMenu *menu, const QString &prefix)
{
    // This function is useful only if the enable action button is active
    QList<QAction*> ret;
    if(enableActions_) {
        foreach (QAction *action, menu->actions()) {
            if (QMenu *submenu = action->menu()) {
                //Flatten hierarchy and prefix submenu text to all actions in submenu
                ret << parseMenu(submenu, action->text());
            } 
            else if (!action->isSeparator()) {
                QString text;
                if (prefix.isEmpty()) {
                    text = action->text().replace(QRegExp("&([\\S])"), "\\1");
                }
                else {
                    text = QString("%1: %2").arg(prefix).arg(action->text()).replace(QRegExp("&([\\S])"), "\\1");
                }

                if (actionsAdded_<=MAX_ADDED_ACTIONS) {
                    actionsAdded_++;
                }
                else {
                    return ret;
                }
                if(actionsAdded_!=1){
	            qDebug() << "Adding action" << text;

		    QAction *a = new QAction(action->icon(), text, this);
		    //KonqMenuActions has the slots that perform the correct actions
		    //so we need our QueryActions to trigger the actions created by
		    //KonqMenuActions
		    connect(a, SIGNAL(triggered(bool)), action, SIGNAL(triggered(bool)));
		    ret << a;
                }
            }
        }
    }
    return ret;
}

// The following code is mostly shared with the nepomuksearch runner and written by Ryan P. Bitanga <ryan.bitanga@gmail.com>.
QList<QAction*> FSRunner::actionsForMatch(const Plasma::QueryMatch &match)
{

    // This function is useful only if the enable action button is active
    QList<QAction*> ret;
    if(enableActions_) {
        QString path = match.data().toString();
        qDebug() << "Actions on path:" << path;

        QString address = QUrl::fromPercentEncoding(path.toUtf8());
        KUrl url(address);
        KIO::UDSEntry entry;
        if (!KIO::NetAccess::stat(url.path(), entry, 0)) {
             return QList<QAction*>();
        }

        KFileItemList list;
        list << KFileItem(entry, url);

        KonqPopupMenuInformation info;
        info.setItems(list);

        QMenu dummy;
        assert(actions_);
        actions_->setPopupMenuInfo(info);
        actions_->addOpenWithActionsTo(&dummy, QString());
        //Add user defined actions
        //Uncomment this if you want the user actions (I don't think it's useful)
        //actions_->addActionsTo(&dummy);
        actionsAdded_=0;

        ret << parseMenu(&dummy);
    }
    return ret;
}

#include "fs.moc"
