/*
 *   Copyright (C) 2008-2009 Anders Aagaard <aagaande@gmail.com
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License version 2 as
 *   published by the Free Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "fs_config.h"

#include <QGridLayout>

#include <KConfigGroup>
#include <KDebug>
#include <KPluginFactory>
#include <KPluginLoader>
#include <QPushButton>
#include <QDialogButtonBox>
#include <KIcon>
#include <cassert>
#include <QSqlTableModel>

#include <plasma/abstractrunner.h>

K_EXPORT_RUNNER_CONFIG(fs, FSConfig)

FsConfigForm::FsConfigForm(QWidget* parent) : QWidget(parent)
{
  setupUi(this);
}

static void setHeaderItem(QStandardItemModel *model, int column, QString title, QString toolTip)
{
    QStandardItem *item = new QStandardItem(title);
    item->setToolTip(toolTip);
    model->setHorizontalHeaderItem(column, item);
}

FSConfig::FSConfig(QWidget* parent, const QVariantList& args) :
        KCModule(ConfigFactory::componentData(), parent, args)
{

    ui = new FsConfigForm(this);

    QGridLayout* layout = new QGridLayout(this);
    layout->addWidget(ui, 0, 0);

    model = new QStandardItemModel(0, 5, this);
    setHeaderItem(model, 0, "Path", QString());
    setHeaderItem(model, 1, "Depth", "How many subdirectories to scan under the path.");
    setHeaderItem(model, 2, "Priority", "Higher priority items will be sorted earlier when searching.");
    setHeaderItem(model, 3, "Files", "Allow searching for files, not only directories.");
    setHeaderItem(model, 4, "Slow", "Insert small delays between folder scans for slow filesystems.");
    connect(model, SIGNAL(itemChanged(QStandardItem*)), SLOT(modelItemChanged(QStandardItem*)));

    assert(ui->treeView);
    ui->treeView->setModel(model);

    QAbstractButton *myAddButton;
    QAbstractButton *myDeleteButton;
    myAddButton = ui->directoriesButtonBox->addButton(tr("Add"), QDialogButtonBox::AcceptRole);
    myAddButton->setIcon(KIcon("document-new"));
    connect(myAddButton, SIGNAL(clicked()), SLOT(addClicked()));

    myDeleteButton = ui->directoriesButtonBox->addButton(tr("Delete"), QDialogButtonBox::RejectRole);
    myDeleteButton->setIcon(KIcon("edit-delete"));
    connect(myDeleteButton, SIGNAL(clicked()), SLOT(deleteClicked()));

    QStringList searchTerms;
    searchTerms << "<Search term>*";
    searchTerms << "*<Search term>*";
    ui->searchTermComboBox->insertItems(0, searchTerms);

    QStringList searchFields;
    searchFields << "Name";
    searchFields << "Full path";
    ui->searchFieldComboBox->insertItems(0, searchFields);


    connect(ui->treatSpaceAsStarCheckbox, SIGNAL(stateChanged(int)), SLOT(changed()));
    connect(ui->enableActionsCheckbox, SIGNAL(stateChanged(int)),SLOT(changed()));
    assert(connect(ui->searchTermComboBox, SIGNAL(currentIndexChanged(int)), SLOT(changed())));
    connect(ui->searchFieldComboBox, SIGNAL(currentIndexChanged(int)), SLOT(changed()));

    load();
}

FSConfig::~FSConfig()
{
}

void FSConfig::modelItemChanged(QStandardItem *item)
{
    QModelIndex idx = model->indexFromItem(item);
    qDebug() << "IDX:" << idx.row() << idx.column() << folders.size();
    assert(idx.row() < folders.size());

    switch (idx.column()) {
        case 0:
        folders[idx.row()].path = item->data(Qt::DisplayRole).toString();
        break;
        case 1:
        folders[idx.row()].subdirs = item->data(Qt::DisplayRole).toInt();
        break;
        case 2:
        folders[idx.row()].priority = item->data(Qt::DisplayRole).toInt();
        break;
        case 3:
        folders[idx.row()].files = item->data(Qt::DisplayRole).toBool();
        break;
        case 4:
        folders[idx.row()].slow = item->data(Qt::DisplayRole).toBool();
        break;
        default:
        qFatal("Couldn't resolve index %d", idx.row());
    }
#ifdef DEBUG
    Folder f = folders[idx.row()];
    qDebug() << "New data : " << f.path << f.subdirs << f.priority << f.files;
#endif
    refresh();
    emit changed(true);
}


void FSConfig::addClicked()
{
    KUrl url;
    QString caption = "Select directory";
    //QString directory = KFileDialog::getExistingDirectory(QDir::currentPath(), this);
    QString directory = KFileDialog::getExistingDirectory(QDir::homePath(), this);
    if (!directory.isEmpty()) {
        for (int i = 0; i < folders.size(); i++)
        {
            if (folders.at(i).path == directory)
                return;
        }
        Folder f;
        f.path = directory;
        f.subdirs = 3;
        f.priority = 0;
        f.files = true;
        f.slow = false;
        folders.append(f);
    }
    refresh();
    emit changed(true);
}

void FSConfig::deleteClicked()
{
    disconnect(model, SIGNAL(itemChanged(QStandardItem*)), this, SLOT(modelItemChanged(QStandardItem*)));
    QModelIndexList selectedList = ui->treeView->selectionModel()->selectedRows();
    if (selectedList.size() == 0)
        return;
    assert(selectedList.size() == 1);
    QModelIndex selected = selectedList.at(0);
    folders.removeAt(selected.row());
    connect(model, SIGNAL(itemChanged(QStandardItem*)), this, SLOT(modelItemChanged(QStandardItem*)));
    refresh();
    emit changed(true);
}


void FSConfig::refresh()
{
    disconnect(model, SIGNAL(itemChanged(QStandardItem*)), this, SLOT(modelItemChanged(QStandardItem*)));
    model->removeRows(0, model->rowCount());
    for (int i = 0; i < folders.size(); i++)
    {
        model->insertRow(i);
        model->setData(model->index(i, 0), folders.at(i).path);
        model->setData(model->index(i, 1), folders.at(i).subdirs);
        model->setData(model->index(i, 2), folders.at(i).priority);
        model->setData(model->index(i, 3), folders.at(i).files);
        model->setData(model->index(i, 4), folders.at(i).slow);
    }

    for (int i = 0; i < model->columnCount(); i++)
        ui->treeView->resizeColumnToContents(i);
    connect(model, SIGNAL(itemChanged(QStandardItem*)), this, SLOT(modelItemChanged(QStandardItem*)));
}

void FSConfig::load()
{
    KCModule::load();
    folders.clear();

    FSSettings settings = loadConfig();
    folders = settings.folders;
    ui->treatSpaceAsStarCheckbox->setChecked(settings.treatSpaceAsStar);
    ui->enableActionsCheckbox->setChecked(settings.enableActions);

    if (!settings.searchTermStrict)
        ui->searchTermComboBox->setCurrentIndex(1);

    if (!settings.searchFieldName)
        ui->searchFieldComboBox->setCurrentIndex(1);

    refresh();
    emit changed(false);
}

void FSConfig::save()
{
    qDebug() << "FSConfig::save()";
    FSSettings settings;
    settings.folders = folders;
    settings.treatSpaceAsStar = ui->treatSpaceAsStarCheckbox->isChecked();
    settings.enableActions = ui->enableActionsCheckbox->isChecked();

    settings.searchTermStrict = ui->searchTermComboBox->currentIndex() == 0;
    settings.searchFieldName = ui->searchFieldComboBox->currentIndex() == 0;

    qDebug() << "Saving settings" << settings.treatSpaceAsStar << settings.searchTermStrict << settings.searchFieldName << settings.enableActions;

    saveSettings(settings);

    emit changed(false);
}

void FSConfig::defaults()
{
    folders.clear();
    refresh();
    ui->treatSpaceAsStarCheckbox->setChecked(true);
    ui->enableActionsCheckbox->setChecked(false);
    ui->searchTermComboBox->setCurrentIndex(0);
    ui->searchFieldComboBox->setCurrentIndex(0);
    emit changed(true);
}


#include "fs_config.moc"
